/*
 * Part of the GaBIEn project.
 * (Graphics and Basic Input Engine)
 *
 *        DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *                    Version 2, December 2004 
 *
 * Modified by gamemanj <asdd2808@gmail.com> to remove expletives
 * Original copyright (C) 2004 Sam Hocevar <sam@hocevar.net> 
 *
 * Everyone is permitted to copy and distribute verbatim or modified 
 * copies of this license document, and changing it is allowed as long 
 * as the name is changed. 
 *
 *            DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION 
 *
 *  0. You just DO WHATEVER YOU WANT TO.
 */
package gabien.music;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

public class PiyoPiyoFile {
    public PiyoPiyoWaveTrack[] waveTracks = new PiyoPiyoWaveTrack[3];
    public int musicWait, drumVolume, loopStart, loopEnd;
    public PiyoPiyoFrame[] musicFrames;

    public PiyoPiyoFile(int musicLength) {
        musicFrames = new PiyoPiyoFrame[musicLength * 4];
        for (int i = 0; i < musicFrames.length; i++)
            musicFrames[i] = new PiyoPiyoFrame();
        waveTracks[0] = new PiyoPiyoWaveTrack();
        waveTracks[1] = new PiyoPiyoWaveTrack();
        waveTracks[2] = new PiyoPiyoWaveTrack();
    }

    // For optimization/expansion tasks.
    // (This is used in the editor so you never run out of space)
    public void resize(int targetLength) {
        PiyoPiyoFrame[] oldMusicFrames = musicFrames;
        musicFrames = new PiyoPiyoFrame[targetLength * 4];
        int omfSection = (oldMusicFrames.length / 4);
        int nmfSection = targetLength;
        for (int i = 0; i < targetLength; i++) {
            if (i < omfSection) {
                musicFrames[i] = oldMusicFrames[i];
                musicFrames[i + nmfSection] = oldMusicFrames[i + omfSection];
                musicFrames[i + (nmfSection * 2)] = oldMusicFrames[i + (omfSection * 2)];
                musicFrames[i + (nmfSection * 3)] = oldMusicFrames[i + (omfSection * 3)];
            } else {
                musicFrames[i] = new PiyoPiyoFrame();
                musicFrames[i + nmfSection] = new PiyoPiyoFrame();
                musicFrames[i + (nmfSection * 2)] = new PiyoPiyoFrame();
                musicFrames[i + (nmfSection * 3)] = new PiyoPiyoFrame();
            }
        }
    }

    // This is here so if this changes in future stuff doesn't break.
    // For now this value has been determined by random experimentation.
    public double beatTime() {
        // If the result is too long, increase, if the result is too short, decrease
        return musicWait / 905.0d;
    }

    public void loadFile(InputStream is) throws IOException {
        if ((read8(is) != ((int) 'P')) || (read8(is) != ((int) 'M')) || (read8(is) != ((int) 'D')))
            throw new IOException("Not a PMD file!");
        read8(is);// song flags
        int pos = read32(is) - 0x418;
        musicWait = read32(is);
        loopStart = read32(is);
        loopEnd = read32(is);
        musicFrames = new PiyoPiyoFrame[read32(is) * 4];
        waveTracks[0].read(is);
        waveTracks[1].read(is);
        waveTracks[2].read(is);
        drumVolume = read32(is);
        is.skip(pos);
        for (int i = 0; i < musicFrames.length; i++) {
            musicFrames[i] = new PiyoPiyoFrame();
            musicFrames[i].read(is);
        }
    }

    public void saveFile(OutputStream os) throws IOException {
        os.write((int)'P');
        os.write((int)'M');
        os.write((int)'D');
        os.write(0x80);
        write32(os, 0x418);
        write32(os, musicWait);
        write32(os, loopStart);
        write32(os, loopEnd);
        write32(os, musicFrames.length/4);
        waveTracks[0].write(os);
        waveTracks[1].write(os);
        waveTracks[2].write(os);
        write32(os, drumVolume);
        for (int i = 0; i < musicFrames.length; i++)
            musicFrames[i].write(os);
    }

    protected static int read8(InputStream is) throws IOException {
        int i = is.read();
        if (i == -1)
            throw new IOException("Unexpected EOF");
        return i;
    }

    protected static int read32(InputStream is) throws IOException {
        int i = read8(is);
        i |= read8(is) << 8;
        i |= read8(is) << 16;
        i |= read8(is) << 24;
        return i;
    }

    protected static void readArray(InputStream is, int[] b) throws IOException {
        for (int i = 0; i < b.length; i++)
            b[i] = read8(is);
    }

    protected static void write32(OutputStream is, int i) throws IOException {
        is.write(i & 0xFF);
        is.write((i >> 8) & 0xFF);
        is.write((i >> 16) & 0xFF);
        is.write((i >> 24) & 0xFF);
    }

    protected static void writeArray(OutputStream is, int[] b) throws IOException {
        for (int i = 0; i < b.length; i++)
            is.write(b[i]);
    }
}
