/*
 * Part of the GaBIEn project.
 * (Graphics and Basic Input Engine)
 *
 *        DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *                    Version 2, December 2004 
 *
 * Modified by gamemanj <asdd2808@gmail.com> to remove expletives
 * Original copyright (C) 2004 Sam Hocevar <sam@hocevar.net> 
 *
 * Everyone is permitted to copy and distribute verbatim or modified 
 * copies of this license document, and changing it is allowed as long 
 * as the name is changed. 
 *
 *            DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION 
 *
 *  0. You just DO WHATEVER YOU WANT TO.
 */
package gabien.music;

import gabien.ISoundDriver;

import java.io.DataOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.LinkedList;

public class PiyoPiyoPlayer {

    // Notes & Samples
    public LinkedList<PiyoPiyoPlayingNote> playingNotes = new LinkedList<PiyoPiyoPlayingNote>();
    public LinkedList<PiyoPiyoPlayingSample> playingSamples = new LinkedList<PiyoPiyoPlayingSample>();

    public double withinBeatPosition = 0;
    public final PiyoPiyoFile piyoFile;
    public ISoundDriver sound;
    public int currentFrame = 0;

    public int[] panValues = new int[4];

    public PiyoPiyoPlayer(FileInputStream fis, ISoundDriver sound)
            throws IOException {
        this.sound = sound;
        piyoFile = new PiyoPiyoFile(1);
        piyoFile.loadFile(fis);
        /*
         * FileOutputStream fos=new FileOutputStream("piyoDrums.bin");
         * DataOutputStream dos=new DataOutputStream(fos); for (int i=0;i<6;i++)
         * { FileInputStream is=new FileInputStream("D"+(i+1)+".raw"); int
         * b=is.available(); dos.writeInt(b); for (int j=0;j<b;j++)
         * dos.write(is.read()); is.close(); } dos.close();
         */
    }

    public PiyoPiyoPlayer(PiyoPiyoFile file, ISoundDriver sound) {
        this.sound = sound;
        piyoFile = file;
        panValues[0] = 3;
        panValues[1] = 3;
        panValues[2] = 3;
        panValues[3] = 3;
    }

    // Update beats. Can be run in reverse.
    public void updateBeats(double deltaTime) {
        withinBeatPosition += deltaTime;
        boolean beat = false;
        int beatFrame = currentFrame;
        double musicWaitTime = piyoFile.beatTime();
        if (withinBeatPosition < 0) {
            beat = true;
            withinBeatPosition += musicWaitTime;
            currentFrame--;
            if (currentFrame < 0) {
                setCurrentFrame(piyoFile.loopEnd);
            }
        }
        if (withinBeatPosition > musicWaitTime) {
            beat = true;
            withinBeatPosition -= musicWaitTime;
            currentFrame++;
            if (((currentFrame * 4) >= piyoFile.musicFrames.length)
                    || (currentFrame >= piyoFile.loopEnd)) {
                setCurrentFrame(piyoFile.loopStart);
            }
        }

        // Now correct the "backwards time"
        if (deltaTime < 0)
            deltaTime = -deltaTime;

        if (beat) {
            for (int i = 0; i < 4; i++)
                handleBeat(piyoFile.musicFrames[beatFrame
                        + ((piyoFile.musicFrames.length / 4) * i)], i);
        }
    }

    // Update envelopes. Cannot be run in reverse.
    public void updateEnvelopes(double deltaTime) {
        for (PiyoPiyoPlayingNote note : new LinkedList<PiyoPiyoPlayingNote>(
                playingNotes)) {
            if (note.updateEnvelope(deltaTime)) {
                sound.deleteChannel(note.channel);
                playingNotes.remove(note);
            }
        }
        for (PiyoPiyoPlayingSample note : new LinkedList<PiyoPiyoPlayingSample>(
                playingSamples)) {
            if (note.doUpdate(deltaTime)) {
                sound.deleteChannel(note.channel);
                playingSamples.remove(note);
            }
        }
    }

    public void setCurrentFrame(int value) {
        currentFrame = value;
        if (currentFrame < 0)
            currentFrame = 0;
        if ((currentFrame * 4) >= piyoFile.musicFrames.length)
            currentFrame = (piyoFile.musicFrames.length / 4) - 1;
    }

    private void handleBeat(PiyoPiyoFrame piyoPiyoFrame, int track) {
        if (piyoPiyoFrame.panValue != 0)
            panValues[track] = piyoPiyoFrame.panValue - 1;
        for (int i = 0; i < 24; i++) {
            if (piyoPiyoFrame.hitNotes[i]) {
                playNote(track, i, panValues[track]);
            }
        }
    }

    public void playNote(int track, int i, int panValue) {
        switch (track) {
        case 0:
        case 1:
        case 2:
            playingNotes.add(new PiyoPiyoPlayingNote(sound.createChannel(),
                    piyoFile.waveTracks[track], i, panValue));
            break;
        case 3:
            playingSamples.add(new PiyoPiyoPlayingSample(sound.createChannel(),
                    i, piyoFile.drumVolume, panValue));
            break;
        default:

            break;
        }
    }

    public void stopAll() {
        for (PiyoPiyoPlayingNote note : playingNotes) {
            sound.deleteChannel(note.channel);
        }
        for (PiyoPiyoPlayingSample note : playingSamples) {
            sound.deleteChannel(note.channel);
        }
        playingNotes.clear();
        playingSamples.clear();
    }
}
