/*
 * Part of the JavaChan project.
 * (Java ikaChan)
 *
 *        DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *                    Version 2, December 2004 
 *
 * Modified by gamemanj <asdd2808@gmail.com> to remove expletives
 * Original copyright (C) 2004 Sam Hocevar <sam@hocevar.net> 
 *
 * Everyone is permitted to copy and distribute verbatim or modified 
 * copies of this license document, and changing it is allowed as long 
 * as the name is changed. 
 *
 *            DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION 
 *
 *  0. You just DO WHATEVER YOU WANT TO.
 */
package javachan;

import gabien.IGrInDriver;
import gabien.ui.Rect;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;

import javachan.npcs.*;

/**
 *
 * @author matt
 */
public class GameStateNPChar {
    public NPCCharacter[] npcTable = new NPCCharacter[100];
    public BasicNPCType[] npcTypes=new BasicNPCType[] {
        new GlobefishNPCType(),
        new IsoginNPCType(),
        new CrabNPCType(),
        new SleepNPCType(),
        new AttackStarNPCType(true),
        new AttackStarNPCType(false),
        new DumNPCType(),
        new CarryNPCType(),
        new JuelNPCType(),
        new UfoNPCType()};
    public static class NPCCharacter implements IPhysicsObject {
        public boolean exists=false;
        public double posX, posY, velX, velY, ofsX, ofsY;
        public int entityType, collisionType, entityStatus, eventID;//collisiontype:0:NULL 1:BLOCK 2:ENEMY 3:EVENT
        // These aren't saved.
        public boolean onGround=false;
        public int entityStatus2, entityDamage, displayX, displayY, touchWillHurt;
        public NPCCharacter() {
        }

        @Override
        public Rect getRectangle() {
            return new Rect((int) (posX * 16), (int) (posY * 16), 16, 16);
        }
        public boolean stopsIkachan() {
            return (collisionType == 3) || (collisionType == 2);
        }
        //Called once per frame at most on collision with Ikachan.
        //Time is the current 'physical time'-essentially the amount of seconds not spent in a script.
        //This is used to avoid Ikachan continually activating the same script.
        public void handleIkachanCollision(GameState gs,IEventCallback iec, double Time) {
            //TODO:Anything involving Ironhead.
            //TODO:HandleIkachanCollision for enemies.
            if (collisionType == 0) {
                if (gs.lastAnnoyTime + 1 <= Time) {
                    iec.startEvent(eventID);
                    gs.lastAnnoyTime = Time;
                }
            }
            if (collisionType == 3) {
                if (gs.lastAnnoyTime + 1 <= Time) {
                    iec.startEvent(eventID);
                    gs.lastAnnoyTime = Time;
                }
            }
        }

        @Override
        public double getX() {
            return posX;
        }

        @Override
        public double getY() {
            return posY;
        }

        @Override
        public double getVX() {
            return velX;
        }

        @Override
        public double getVY() {
            return velY;
       }

        @Override
        public void setX(double X) {
            this.posX=X;
        }

        @Override
        public void setY(double Y) {
            this.posY=Y;
        }

        @Override
        public void setVX(double VX) {
            this.velX=VX;
        }

        @Override
        public void setVY(double VY) {
            this.velY=VY;
        }
    }

    // Apparently, negative values actually do work in this function!
    private static void doubleToFixedPoint(double d, FileOutputStream fos) throws IOException {
        int dd=(int)Math.floor(d);
        dd=dd<<14;
        double fractional=d-Math.floor(d);
        fractional*=1<<14;
        dd+=(int) fractional;
        fos.write((dd&0xFF));
        fos.write((dd&0xFF00)>>8);
        fos.write((dd&0xFF0000)>>16);
        fos.write((dd&0xFF000000)>>24);
    }

    private static double fixedPointToDouble(InputStream is) throws IOException {
        // But not in the old version of this one.
        /*
        double fractional = is.read();
        int mb = is.read();//WWFFFFFF
        fractional += (mb & 0x3F) << 8;
        fractional /= (1 << 14);
        int whole = (mb & (64 | 128)) / 64;
        whole |= is.read() << 2;
        whole |= is.read() << 10;
        return fractional + whole;*/
        int i=is.read();
        i|=is.read()<<8;
        i|=is.read()<<16;
        i|=is.read()<<24;
        return i/((double)(1<<14));
    }

    
    public GameStateNPChar()
    {
        for (int p = 0; p < npcTable.length; p++) {
            npcTable[p]=new NPCCharacter();
        }
    }
    
    
    public void load(InputStream is) throws Exception {
        for (int p = 0; p < npcTable.length; p++) {
            NPCCharacter nc = new NPCCharacter();
            boolean sw = is.read() != 0;
            nc.collisionType = is.read();
            nc.entityType = is.read();
            nc.eventID = is.read();
            nc.entityStatus=is.read();//?
            nc.entityStatus |= is.read() << 8;
            nc.posX = fixedPointToDouble(is);
            nc.posY = fixedPointToDouble(is);
            nc.ofsX = fixedPointToDouble(is);
            nc.ofsY = fixedPointToDouble(is);
            nc.exists=sw;
            // initialize unsaveds
            nc.displayX = 0;
            nc.displayY = 0;
            nc.onGround=false;
            nc.entityDamage=0;
            nc.touchWillHurt=0;
            nc.entityStatus2=0;
            
            npcTable[p] = nc;
        }
    }
   
    public void save(FileOutputStream fos) throws IOException {
        for (int p = 0; p < npcTable.length; p++) {
            NPCCharacter nc=npcTable[p];
            if (nc.exists)
            { 
                fos.write(1);
            } else {
                fos.write(0);
            }
            fos.write(nc.collisionType);
            fos.write(nc.entityType);
            fos.write(nc.eventID);
            fos.write(nc.entityStatus&0xFF);
            fos.write((nc.entityStatus&0xFF00)>>8);
            doubleToFixedPoint(nc.posX,fos);
            doubleToFixedPoint(nc.posY,fos);
            doubleToFixedPoint(nc.ofsX,fos);
            doubleToFixedPoint(nc.ofsY,fos);
        }
    }

    public void doRender(int px, int py, double Time,IGrInDriver gd) {
        for (int p = 0; p < npcTable.length; p++)
            if (npcTable[p].exists)
                npcTypes[npcTable[p].entityType].doRender(px, py,Time,npcTable[p],gd);
    }
    
    public void doEditorTick()
    {
        for (int p = 0; p < npcTable.length; p++)
            if (npcTable[p].exists)
                npcTypes[npcTable[p].entityType].doEditorTick(npcTable[p]);
    }
}
