/*
 * Part of the JavaChan project.
 * (Java ikaChan)
 *
 *        DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *                    Version 2, December 2004 
 *
 * Modified by gamemanj <asdd2808@gmail.com> to remove expletives
 * Original copyright (C) 2004 Sam Hocevar <sam@hocevar.net> 
 *
 * Everyone is permitted to copy and distribute verbatim or modified 
 * copies of this license document, and changing it is allowed as long 
 * as the name is changed. 
 *
 *            DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION 
 *
 *  0. You just DO WHATEVER YOU WANT TO.
 */
package javachan.games;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.logging.Level;
import java.util.logging.Logger;

import javachan.Game;
import javachan.GameState;
import javachan.IEventCallback;
import javachan.IGame;
import javachan.PhysicsEngine;
import gabien.GaBIEn;
import gabien.IGrInDriver;
import gabien.IGrInDriver.IImage;
import javachan.games.minorstates.*;

/**
 * Uh, the main game... (including a TSC interpreter and such)
 * 
 * @author matt
 */
public class MainGame implements IGame, IEventCallback {
    public static final double tickLen = 0.02;
    public double dcamX = 0, dcamY = 0;
    public GameState gameState = new GameState();
    public int scriptPoint;
    public String currentChat = null;
    public int currentChatReveal = 0;
    // This is where all the TSC-related Minor States go.
    public ChatMinorState chatMinorStateInstance = new ChatMinorState();
    public String chatState = "";
    public boolean showMap = false;
    public double tickTimer = 0;
    public double fadeTimer = 0;// Fades apparently don't pause the script.
    public boolean thisIsEndingFade = false;
    public boolean returningFromMaru = false;
    public int fadeState = 0;
    public int savedLR = 0;// used for displaying Ikachan
    public PhysicsEngine physicsEngine = new PhysicsEngine();
    // SCRIPT STATE VARIABLES
    public int lastEventLookup = 0;// Last Event Lookup
    public int selectedItem = 0;
    public int majorState = 0;// 0:Startup,1:Main Game,2:Dead
    // All MinorStates except null mean paused.
    public MinorState minorState = null;

    // END SCRIPT STATE VARIABLES
    public boolean runScriptCommand(double Delta, Game g) {
        String s = getLine(scriptPoint++);
        System.out.println("Script line execution:" + s);
        if (s.startsWith("<")) {
            // COMMANDS
            if (s.startsWith("<e+")) {
                gameState.playerEnergy = gameState.getMaxEnergy();
                return false;
            }
            if (s.startsWith("<ex")) {
                System.exit(0);
                return false;
            }
            if (s.startsWith("<re")) {
                gameState.restoreCharacterPosition();
                return false;
            }
            if (s.startsWith("<yn")) {
                int yesDest = lookupEvent(parseInt(s.substring(3))) + 1;
                minorState = new YesNoMinorState(yesDest);
                g.graphics.clearKeys();
                return false;
            }
            if (s.startsWith("<ge")) {
                gameState.initGreatEscape();
                return true;
            }
            if (s.startsWith("<bo")) {
                gameState.bossMode = true;
                gameState.bosses[gameState.bossID].initBattle();
                gameState.bossID = parseInt(s.substring(3));
                return false;
            }
            if (s.startsWith("<pd")) {
                String r = "Pmd/Tidepool.pmd";
                int i = parseInt(s.substring(3));
                if (i == 0) {
                    r = "Pmd/Ikachan.pmd";
                }
                if (i == 1) {
                    r = "Pmd/Magirete.pmd";
                }
                if (i == 2) {
                    r = "Pmd/Buriki.pmd";
                }
                if (i == 3) {
                    r = "Pmd/Mizuno.pmd";
                }
                if (i == 4) {
                    r = "Pmd/Quake.pmd";
                }
                g.music.setMusic(r);
                g.music.playMusic();
                return true;
            }
            if (s.startsWith("<pp")) {
                g.music.playMusic();
                return true;
            }
            if (s.startsWith("<ps")) {
                g.music.stopMusic();
                return true;
            }
            if (s.startsWith("<fi")) {
                fadeState = 1;
                fadeTimer = 0;
                return true;
            }
            if (s.startsWith("<fo")) {
                fadeState = 2;
                fadeTimer = 0;
                return true;
            }
            if (s.startsWith("<ec")) {
                gameState.entityCount++;
                return true;
            }
            if (s.startsWith("<fj")) {
                if (gameState.flags[parseInt(s.substring(3))])
                    scriptPoint = lookupEvent(parseInt(s.substring(8))) + 1;
                return true;
            }
            if (s.startsWith("<ij")) {
                int pi = parseInt(s.substring(3));
                for (int xx = 0; xx < gameState.items.length; xx++) {
                    if (gameState.items[xx] == pi) {
                        scriptPoint = lookupEvent(parseInt(s.substring(8))) + 1;
                        break;// Welp,we're done here.
                    }
                }
                return true;
            }
            if (s.startsWith("<ej")) {
                if (parseInt(s.substring(3)) == gameState.entityCount)
                    scriptPoint = lookupEvent(parseInt(s.substring(8))) + 1;
                return true;
            }
            if (s.startsWith("<nm")) {
                showMap = true;
                return true;
            }
            if (s.startsWith("<ma")) {
                showMap = false;
                return true;
            }
            if (s.startsWith("<i+")) {
                for (int xx = 0; xx < gameState.items.length; xx++) {
                    if (gameState.items[xx] == 0) {
                        gameState.items[xx] = parseInt(s.substring(3));
                        break;// mystery solved!
                    }
                }
                return true;
            }
            if (s.startsWith("<i-")) {
                int pi = parseInt(s.substring(3));
                for (int xx = 0; xx < gameState.items.length; xx++) {
                    if (gameState.items[xx] == pi) {
                        gameState.items[xx] = 0;
                        break;
                    }
                }
                return true;
            }
            if (s.startsWith("<f+")) {
                gameState.flags[parseInt(s.substring(3))] = true;
                return true;
            }
            if (s.startsWith("<cm")) {
                int npcc = parseInt(s.substring(3));
                int npcx = parseInt(s.substring(3 + 5));
                int npcy = parseInt(s.substring(3 + 5 + 5));
                gameState.npcs.npcTable[npcc].posX = npcx;
                gameState.npcs.npcTable[npcc].posY = npcy;
                if (npcx == 0)
                    if (npcy == 0)
                        gameState.npcs.npcTable[npcc].exists = false;
                return true;
            }
            if (s.startsWith("<wa")) {
                int pca = parseInt(s.substring(3));
                currentChat = null;
                chatState = "";
                chatMinorStateInstance.generalMinorUseTimer = pca * tickLen;
                updateChatMinorState(5);
                return false;
            }
            if (s.startsWith("<pc")) {
                int pca = parseInt(s.substring(3));
                int pcb = parseInt(s.substring(3 + 5));
                int pcc = parseInt(s.substring(3 + 5 + 5));
                gameState.map.data[pca + (pcb * gameState.map.width)] = pcc;
                return true;
            }
            if (s.startsWith("<end")) {
                if (majorState == 0) {
                    majorState = 1;
                    scriptPoint = lookupEvent(2) + 1;
                    g.graphics.clearKeys();
                    g.music.stopMusic();
                    if (g.graphics.isKeyDown(Keymap.VK_ITEMS)) {
                        g.gameState = new EditorGame(gameState);
                    } else {
                        g.music.setMusic("Pmd/Magirete.pmd");
                        g.music.playMusic();
                        returningFromMaru = true;
                        g.gameState = new EndingGame(true, this);
                    }
                } else if (majorState == 2) {
                    scriptPoint = lookupEvent(6) + 1;
                    majorState = 1;
                } else {
                    scriptPoint = -1;
                }
                return false;
            }
            System.out.println("NYI:" + s);
            return true;
        } else {

            // Note that { is used by Japanese versions of Ikachan.
            if (s.endsWith("/")||s.endsWith("{")) {
                s = s.substring(0, s.length() - 1);
                updateChatMinorState(1);
            } else {
                updateChatMinorState(4);
            }
            String LastChatState = chatState;
            if (s.endsWith("+")) {
                s = s.substring(0, s.length() - 1);
                chatState = "+";
            } else {
                chatState = "";
            }
            if (LastChatState.equals("+")) {
                currentChat += "\r\n" + s;
            } else {
                if (minorState instanceof ChatMinorState) {
                    ((ChatMinorState) minorState).generalMinorUseTimer = 0;
                }
                currentChat = s;
                currentChatReveal = 0;
            }
            return false;// needs to get on the screen :)
        }
    }

    private void updateChatMinorState(int i) {
        if (!(minorState instanceof ChatMinorState)) {
            minorState = chatMinorStateInstance;
        }
        ((ChatMinorState) minorState).subState = i;
    }

    public MainGame() {
        startEvent(1);
        try {
            gameState.map.loadBitmap(GaBIEn.getFile("Pbm/Map1.pbm"));
            gameState.npcs.load(GaBIEn.getFile("NPChar.dat"));
        } catch (Exception ex) {
            Logger.getLogger(MainGame.class.getName()).log(Level.SEVERE,
                    "Couldn't load main map!!!", ex);
        }
    }

    // This function is structured so the JustPress functions do only get called
    // once.
    // Just look carefully at the code paths...(really really carefully!)
    @Override
    public void doUpdate(double deltaTime, double Time, Game g) {
        if (returningFromMaru) {
            returningFromMaru = false;
            g.music.setMusic("Pmd/Ikachan.pmd");
        }
        /*
         * CHEAT if (g.graphics.getMouseJustDown()) { gameState.playerLevel=6;
         * gameState.playerEnergy=gameState.getMaxEnergy();
         * gameState.items[0]=1; gameState.playerX=49; gameState.playerY=35; }
         * if (g.graphics.getMouseJustDown()) { gameState.playerLevel=6;
         * gameState.playerEnergy=gameState.getMaxEnergy();
         * gameState.playerX=67; gameState.playerY=79; }
         */
        tickTimer += deltaTime;
        boolean tickNow = false;
        if (tickTimer > tickLen) {
            tickTimer -= tickLen;
            tickNow = true;
        }

        // --------------------------UPDATE
        if (minorState == null) {
            if (scriptPoint != -1) {
                // Run until signal given 'done for this frame'
                while (runScriptCommand(deltaTime, g)) {
                }
            }
        }
        // now,after that is done,re-test
        if (scriptPoint == -1) {
            if (gameState.playerDead) {
                majorState = 2;
                scriptPoint = -5;
            }
            currentChat = null;// done with chat,no script running==no chat!
            chatState = "";
            minorState = null;// Since we've just Nulled out CurrentChat,better
                              // turn off minorstates
            // PHYSICS
            gameState.inGameTime += deltaTime;
            int LR = 0;
            if (g.graphics.isKeyDown(Keymap.VK_LEFT))
                LR = -1;
            if (g.graphics.isKeyDown(Keymap.VK_RIGHT))
                LR = 1;
            // During normal gameplay,only grab Swim on tick boundaries.
            // During scripting,grab it at all times because we're getting
            // called all the time
            // (instead of on tick boundaries.)

            // Make sure these buffers are 0,lest we go into a minorstate
            // unprepared :)
            // We can't use clearkeys as that would break Swim.
            g.graphics.isKeyJustPressed(Keymap.VK_LEFT);
            g.graphics.isKeyJustPressed(Keymap.VK_RIGHT);
            if (gameState.greatEscape)
                LR = 0;
            if (tickNow) {
                gameState.particles.doUpdate();
                if (gameState.greatEscape) {
                    gameState.greatEscapeTick();
                    gameState.mainNPCTick(physicsEngine, gameState.inGameTime);
                } else {
                    gameState.bossesTick();
                    gameState.mainTick(LR,
                            g.graphics.isKeyJustPressed(Keymap.VK_SWIM),
                            g.graphics.isKeyDown(Keymap.VK_SWIM),
                            physicsEngine, (IEventCallback) this,
                            gameState.inGameTime);
                    gameState.mainNPCTick(physicsEngine, gameState.inGameTime);
                }
                if (gameState.playerY < 0) {
                    if (fadeState == 0) {
                        if (thisIsEndingFade) {
                            g.music.stopMusic();
                            g.music.setMusic("Pmd/Tidepool.pmd");
                            g.music.playMusic();
                            g.graphics.clearKeys();
                            g.gameState = new EndingGame(false,
                                    new OpeningGame());
                            // just do nothing-graphics won't flush,so it'll
                            // look fine
                            return;
                        } else {
                            fadeState = 2;
                            fadeTimer = 0;
                            thisIsEndingFade = true;
                        }
                    } else {
                    }
                }
            }
            if (g.graphics.isKeyJustPressed(Keymap.VK_ITEMS)) {
                // If we're not in a script(physics could relatively easily
                // cause a problem)...
                if (scriptPoint == -1) {
                    scriptPoint = -2;
                    minorState = new ItemsMinorState();
                    selectedItem = 0;
                }
            }

            savedLR = LR;
        }

        // --------------------------DRAWING

        g.graphics.clearAll(0, 0, 0);
        // draw map here,ikachan included at no extra cost.
        // RENDERING.Call in this order:RenderBack,Map.Render(pass
        // 0),NPChar.Render,RenderPC,Map.Render(pass 1)
        double tCamX = -((gameState.playerX * 16) - (g.graphics.getWidth() / 2));
        double tCamY = -((gameState.playerY * 16) - (g.graphics.getHeight() / 2));
        dcamX += ((tCamX - dcamX) / 64) * 1;
        dcamY += ((tCamY - dcamY) / 64) * 1;
        int camX = (int) dcamX;
        int camY = (int) dcamY;
        // note: it doesn't do the adjustments on the lower-right corner. just a
        // visual problem
        if (camX > 0) {
            camX = 0;
        }
        if (camY > 0) {
            camY = 0;
        }
        if (showMap) {
            gameState.renderBack(camX, camY, Time, g.graphics);
            gameState.map.renderTiles(camX, camY, 0, Time, g.graphics);
            gameState.npcs.doRender(camX, camY, gameState.inGameTime,
                    g.graphics);
            gameState.renderPC(camX, camY, savedLR, g.graphics,
                    gameState.inGameTime);
            gameState.map.renderTiles(camX, camY, 1, Time, g.graphics);
            gameState.particles.render(camX, camY, g.graphics);
        }
        if (currentChat != null) {
            IImage chatbox = g.graphics.getImage("Pbm/Msgbox.pbm", 0, 0, 0);
            String c = currentChat.substring(0, currentChatReveal);
            g.graphics
                    .blitBCKImage(0, 0, chatbox.getWidth(),
                            chatbox.getHeight(), 10, g.graphics.getHeight()
                                    - (chatbox.getHeight() + 9), chatbox);
            if (c.length() > 0) {
                String co = c;
                if (co.indexOf("\r\n") != -1) {
                    co = co.substring(0, co.indexOf("\r\n"));
                }
                g.graphics.drawText(20,
                        (g.graphics.getHeight() + 2)
                                - (chatbox.getHeight() + 1), 0, 0, 255, 12, co);
                g.graphics.drawText(18,
                        g.graphics.getHeight() - (chatbox.getHeight() + 1),
                        255, 255, 255, 12, co);
                if (c.split("\r\n").length > 1) {
                    g.graphics.drawText(20, (g.graphics.getHeight() + 2)
                            - (chatbox.getHeight() - 11), 0, 0, 255, 12,
                            c.split("\r\n")[1]);
                    g.graphics
                            .drawText(
                                    18,
                                    g.graphics.getHeight()
                                            - (chatbox.getHeight() - 11), 255,
                                    255, 255, 12, c.split("\r\n")[1]);
                }
            }
        }
        if (scriptPoint == -1) {
            IImage statspage = g.graphics.getImage("Pbm/Status.pbm", 0, 0, 0);
            int b = 8;
            int d = 48;
            g.graphics.blitBCKImage(0, 0, statspage.getWidth(),
                    statspage.getHeight(), b, b, statspage);
            drawFigure(g.graphics, gameState.playerLevel, b + d, b);
            drawFigure(g.graphics, gameState.playerXP, b + d, b + 8);
            drawFigure(g.graphics, gameState.playerEnergy, b + d, b + 16);

            drawFigure(g.graphics, gameState.getMaxXP(),
                    b + (statspage.getWidth() - 8), b + 8);
            drawFigure(g.graphics, gameState.getMaxEnergy(),
                    b + (statspage.getWidth() - 8), b + 16);
        }
        // ---------------------------MINOR STATE
        // HANDLERS--------------------------------
        if (minorState != null)
            minorState.doUpdate(this, g, deltaTime, tickNow);
        // ---------------------FADES(GOES ON TOP OF EVERYTHING ELSE)
        if (fadeState == 1) {
            fadeTimer += deltaTime * 32;
            if (fadeDrawer((int) fadeTimer, false, g.graphics))
                fadeState = 0;
        }
        if (fadeState == 2) {
            fadeTimer += deltaTime * 32;
            if (fadeDrawer((int) fadeTimer, true, g.graphics))
                fadeState = 0;
        }
        g.graphics.flush();
    }

    // lines start at 1.
    private String getLine(int line) {

        if (line == -5) {
            return "<wa0080";
        }
        if (line == -4) {
            return "<end";
        }
        if (line == -3) {
            return "[" + lastEventLookup + "] Event Error./";// This is so we
                                                             // can alert the
                                                             // user when some
                                                             // event doesn't
                                                             // exist.
        }
        if (line == -2) {
            return "<end";// Problem is,if a chatline comes before -1,
                          // it will be ignored,because the -1 is a signal to
                          // stop execution.
        }
        try {
            BufferedReader br = new BufferedReader(new InputStreamReader(
                    GaBIEn.getFile("Event.ptx"),"Shift-JIS"));
            String s = "";
            while (line > 0) {
                line--;
                s = br.readLine();
            }
            br.close();
            return s;
        } catch (Exception ex) {
            Logger.getLogger(MainGame.class.getName()).log(Level.SEVERE, null,
                    ex);
        }
        return null;
    }

    private int lookupEvent(int i) {
        int Line = 1;
        lastEventLookup = i;
        while (true) {

            String s = getLine(Line);
            if (s == null) {
                System.out.println("Failed lookup for " + i);
                return -4;
            }
            if (s.startsWith("#e"))
                if (parseInt(s.substring(2, 6)) == i) {
                    return Line;
                }
            Line++;
        }
    }

    private int parseInt(String substring) {
        int A = substring.charAt(3) - '0';
        int B = substring.charAt(2) - '0';
        int C = substring.charAt(1) - '0';
        int D = substring.charAt(0) - '0';
        C += D * 10;
        B += C * 10;
        A += B * 10;
        return A;
    }

    @Override
    public void startEvent(int ev) {
        scriptPoint = lookupEvent(ev) + 1;
    }

    // This is a static routine because it's technically global-fade out is done
    // by Opening
    // before we switch gamestates and do the fade in.
    public static boolean fadeDrawer(int stage, boolean out, IGrInDriver to) {
        IImage fade = to.getImage("Pbm/Fade.pbm", 0, 0, out ? 0 : 16);
        boolean ended = true;
        for (int x = 0; x < to.getWidth(); x += 16)
            for (int y = 0; y < to.getHeight(); y += 16) {
                int lstage = stage;
                int dy = to.getHeight() - y;
                int dx = to.getWidth() - x;
                int p = (dy / 16) + (x / 16);
                lstage -= p;
                if (lstage < 0) {
                    lstage = 0;
                }
                if (lstage > 15) {
                    lstage = 15;
                } else {
                    ended = false;
                }
                to.blitBCKImage(lstage * 16, 0, 16, 16, x, y, fade);
            }
        return ended;
    }

    private void drawFigure(IGrInDriver graphics, int PlayerEnergy, int i,
            int i0) {
        IImage fig = graphics.getImage("Pbm/Figure.pbm", 0, 0, 0);
        String ss = "" + PlayerEnergy;
        int p = 0;
        for (int xx = ss.length() - 1; xx >= 0; xx--) {
            char c = ss.charAt(xx);
            graphics.blitBCKImage((c - '0') * 8, 0, 8, 8, i + p, i0, fig);
            p -= 9;
        }
    }
}
